@@ -7,9 +7,14 @@ import re
 import base64
 import json
 import logging
+import uuid
+
 
 import dsnparse
-from datatypes import Datetime
+from datatypes import (
+    Datetime,
+    property as cachedproperty,
+)
 
 from .compat import *
 from . import utils
@@ -207,7 +212,7 @@ class Schema(object):
 
     @property
     def magic_fields(self):
-        """the magic fields for the schema"""
+        """the magic fields for the schema, magic fields start with an underscore"""
         return {f:v for f, v in self.fields.items() if f.startswith('_')}
 
     @property
@@ -220,6 +225,12 @@ class Schema(object):
             pk_name = None
         return pk_name
 
+    @property
+    def pk_names(self):
+        """Returns all the field names comprising the primary key as a list"""
+        pk_name = self.pk_name
+        return [pk_name] if pk_name else []
+
     @classmethod
     def get_instance(cls, orm_class, **kwargs):
         """return a Schema singleton instance for the given orm_class
@@ -231,34 +242,47 @@ class Schema(object):
         :returns: Schema
         """
         table_name = orm_class.table_name
-        if table_name not in cls.instances:
+        if table_name in cls.instances:
+            s = cls.instances[table_name]
+
+        else:
             s = cls(table_name)
             s.orm_class = orm_class
 
-            seen_properties = set()
+            seen_properties = {}
             for klass in inspect.getmro(orm_class)[:-1]:
                 for k, v in vars(klass).items():
                     k = String(k)
-                    if k not in seen_properties:
-                        if isinstance(v, (Field, Index)):
-                            s.set(k, v)
+                    field = None
 
-                        elif isinstance(v, type):
-                            # We've defined a Field class inline of the Orm, so
-                            # we want to instantiate it and set it in all the places
-                            if issubclass(v, Field):
-                                field = v.get_instance()
-                                s.set(k, field)
+                    if isinstance(v, Field):
+                        field = v
 
-                        seen_properties.add(k)
+                    elif isinstance(v, Index):
+                        s.set_index(k, v)
+                        seen_properties[k] = v
+
+                    elif isinstance(v, type) and issubclass(v, Field):
+                        # We've defined a Field class inline of the Orm, so
+                        # we want to instantiate it and set it in all the places
+                        field = v.get_instance()
+
+                    else:
+                        if v is None:
+                            seen_properties[k] = v
+
+                    if field:
+                        if k in seen_properties:
+                            if seen_properties[k] is None:
+                                for fn in field.names:
+                                    s.lookup["field_names_deleted"][fn] = k
+
+                        else:
+                            s.set_field(k, field)
+                            seen_properties[k] = field
 
-            #s.orm_class = orm_class
             cls.instances[table_name] = s
 
-        else:
-            s = cls.instances[table_name]
-            s.orm_class = orm_class
-
         return s
 
     def __init__(self, table_name, **fields_or_indexes):
@@ -283,8 +307,8 @@ class Schema(object):
         self.table_name = String(table_name)
         self.orm_class = None
         self.lookup = {
-            "names": {},
-            "pk": None,
+            "field_names": {},
+            "field_names_deleted": {},
         }
 
         for name, val in fields_or_indexes.items():
@@ -308,7 +332,7 @@ class Schema(object):
             self.set_index(name, val)
 
         else:
-            raise TypeError("not a Field or Index instance")
+            raise TypeError("Not a Field or Index instance")
 
     def __getattr__(self, name):
         """
@@ -317,22 +341,26 @@ class Schema(object):
         return -- string -- the string value of the attribute name, eg, self.foo returns "foo"
         """
         try:
-            return self.lookup["names"][name]
+            return self.lookup["field_names"][name]
 
         except KeyError:
             raise AttributeError("No {} field in schema {}".format(name, self.table_name))
 
     def has_field(self, field_name):
         """Return True if schema contains field_name"""
-        return field_name in self.lookup["names"]
+        return field_name in self.lookup["field_names"]
+
+    def __contains__(self, field_name):
+        return self.has_field(field_name)
 
     def set_field(self, field_name, field):
         if not field_name: raise ValueError("field_name is empty")
         if field_name in self.fields: raise ValueError("{} already exists and cannot be changed".format(field_name))
         if not isinstance(field, Field): raise ValueError("{} is not a Field instance".format(type(field)))
 
-        field.name = field_name
-        field.orm_class = self.orm_class
+        field.__set_name__(self.orm_class, field_name)
+        #field.name = field_name
+        #field.orm_class = self.orm_class
 
         if field.unique:
             self.set_index(field_name, Index(field_name, unique=True))
@@ -343,11 +371,7 @@ class Schema(object):
         self.fields[field_name] = field
 
         for fn in field.names:
-            self.lookup["names"][fn] = field
-
-        if field.is_pk():
-            self.lookup["pk"] = field
-            self.lookup["names"]["pk"] = field
+            self.lookup["field_names"][fn] = field
 
         return self
 
@@ -365,7 +389,7 @@ class Schema(object):
             raise ValueError("index_name must have a value")
         if index_name in self.indexes:
             raise ValueError("index_name {} has already been defined on {}".format(
-                index_name, str(self.indexes[index_name].fields)
+                index_name, str(self.indexes[index_name].field_names)
             ))
         if not isinstance(index, Index): raise ValueError("{} is not an Index instance".format(type(index)))
 
@@ -375,14 +399,30 @@ class Schema(object):
         self.indexes[index_name] = index
         return self
 
-    def field_name(self, k):
+    def field_name(self, field_name, *default):
         """
-        get the field name of k
+        get the canonical field name of field_name
 
-        most of the time, the field_name of k will just be k, but this makes special
-        allowance for k's like "pk" which will return _id
+        most of the time, the field name of field_name will just be field_name,
+        but this checks the configured aliases to return the canonical name
+
+        :param field_name: str, the field_name you want the canonical field name for
+        :param *default: mixed, if present this will be returned instead of AttributeError
+            raised if field_name doesn't exist
         """
-        return self.__getattr__(k).name
+        try:
+            return self.__getattr__(field_name).name
+
+        except AttributeError:
+            if default:
+                return default[0]
+
+            else:
+                if field_name in self.lookup["field_names_deleted"]:
+                    return self.lookup["field_names_deleted"][field_name]
+
+                else:
+                    raise
 
     def create_orm(self, orm_class=None):
         """If you have a schema but don't have an Orm for it, you can call this method
@@ -398,7 +438,7 @@ class Schema(object):
             orm_class = Orm
 
         child_class = type(
-            ByteString(self.table_name) if is_py2 else String(self.table_name),
+            String(self.table_name),
             (orm_class,),
             {"table_name": self.table_name, "schema": self}
         )
@@ -428,7 +468,7 @@ class Schema(object):
 class Index(object):
     """Each index on the table is configured using this class"""
 
-    def __init__(self, *fields, **options):
+    def __init__(self, *field_names, **options):
         """
         initialize an index
 
@@ -437,11 +477,11 @@ class Index(object):
         **options -- dict --
             unique -- boolean -- True if the index should be unique, false otherwise
         """
-        if not fields:
-            raise ValueError("fields list is empty")
+        if not field_names:
+            raise ValueError("field_names list is empty")
 
         self.name = ""
-        self.fields = list(map(str, fields))
+        self.field_names = list(map(String, field_names))
         self.options = options
         self.unique = options.get("unique", False)
 
@@ -456,58 +496,51 @@ class FieldMeta(type):
     This makes it so an embedded field class will be treated as a descriptor. It
     uses an embedded .instance property to actually perform the operations
     """
-    def __get__(cls, obj, classtype=None):
-        if obj is None:
-            # class is requesting this property, so return it
-            return cls
+    def __set_name__(cls, orm_class, name):
+        """I'm not quite sure why this works, but it does. Basically, when you define
+        the field as an embedded class in the orm, this method will get called
+        when the class is being parsed/loaded the first time.
 
-        ret = cls.get_instance().__get__(obj, classtype)
-        return ret
+        So this method gets called and we use cls to create a Field instance and
+        we then set our newly created instance onto orm_class.name. So after this
+        method is called, orm_class.name will be a Field instance instead of a
+        Field subclass
+        """
+        logger.debug(f"Creating field descriptor {orm_class.__name__}.{name} with class {cls.__name__}")
 
-    def __set__(cls, *args, **kwargs):
-        ret = cls.get_instance().__set__(*args, **kwargs)
-        return ret
-
-    def __delete__(cls, *args, **kwargs):
-        ret = cls.get_instance().__delete__(*args, **kwargs)
-        return ret
+        instance = cls(cls.type, cls.required, cls.options)
+        instance.__set_name__(orm_class, name)
+        setattr(orm_class, name, instance)
 
 
 class Field(object, metaclass=FieldMeta):
     """Each column in the database is configured using this class
 
-    You can set a couple getters and setters on this object in order to fine tune
+    You can set some getters and setters on this object in order to fine tune
     control over how a field in the db is set and fetched from the instance and
     also how it is set and fetched from the interface
 
-    In model.Orm, there are popluate() and depart() class methods, when populate
-    is called, each Field will have its iget() method called. When depart() is called, 
-    each Field will have its iset() method called. By customizing these, you can
-    control functionality when a Field is going to be read or written to the db
-
     to customize the field on set and get from the db, you can use the decorators,
     which are classmethods:
 
+    :Example:
         foo = Field(str, True)
 
         @foo.igetter
         def foo(self, val):
-            # do custom things
+            # do custom things when pulling foo from the database
             return val
 
         @foo.isetter
         def foo(self, val):
-            # do custom things
+            # do custom things when putting foo into the database
             return val
 
-    NOTE -- the iset/iget methods are different than traditional python getters
-    and setters because they always need to return a value and they always take in a
-    value
-
     There are also fget/fset/fdel methods that can be set to customize behavior
     on when a value is set on a particular orm instance, so if you wanted to make sure
     that bar was always an int when it is set, you could:
 
+    :Example:
         bar = Field(int, True)
 
         @bar.fsetter
@@ -524,9 +557,48 @@ class Field(object, metaclass=FieldMeta):
             # convert val to something json safe
             return val
 
-    NOTE -- the fget/fset/fdel methods are different than traditional python getters
-    and setters because they always need to return a value and they always take in a
-    value
+    NOTE -- all these get/set/delete methods are different than traditional python getters
+    and setters because they always need to return a value and they always take in the
+    object instance manipulating them and the value.
+
+    NOTE -- Foreign key's can be passed orm instances from other classes because
+    those classes will call the FK's field methods when getting/setting the field
+
+    You can also configure the Field using a class:
+
+    :Example:
+        class bar(Field):
+            type = int
+            required = True
+            options = {
+                "default": 0,
+                "unique": True,
+                "max_size": 512,
+            }
+
+            def fget(self, orm, v):
+                print("fget")
+                return v
+
+            def iget(self, orm, v):
+                print("iget")
+                return v
+
+            def fset(self, orm, v):
+                print("fset")
+                return v
+
+            def fdel(self, orm, v):
+                print("fdel")
+                return v
+
+            def iquery(self, query, v):
+                print("iquery")
+                return v
+
+            def jsonable(self, orm, v):
+                print("jsonable")
+                return v
 
     https://docs.python.org/2/howto/descriptor.html
     """
@@ -555,25 +627,22 @@ class Field(object, metaclass=FieldMeta):
     index = False
     """True if this field is indexed"""
 
-    instance = None
-    """Don't touch this unless you know what you're doing. Holds internal cached instance"""
+    orm_class = None
+    """Holds the model class this field is defined on"""
 
-    @property
+    @cachedproperty(cached="_schema")
     def schema(self):
         """return the schema instance if this is reference to another table
 
         see .set_type() for an explanation on why we defer figuring this out until now
         """
-        if not hasattr(self, "_schema"):
-            field_type = self.original_type
-            module, klass = utils.get_objects(field_type)
-            schema = klass.schema
-            if not schema:
-                raise ValueError("Field type {} is not an Orm class".format(field_type))
+        field_type = self.original_type
+        module, klass = utils.get_objects(field_type)
+        schema = klass.schema
+        if not schema:
+            raise ValueError("Field type {} is not an Orm class".format(field_type))
 
-            self._schema = schema
-
-        return self._schema
+        return schema
 
     @property
     def type(self):
@@ -610,19 +679,33 @@ class Field(object, metaclass=FieldMeta):
 
     @property
     def names(self):
-        names = []
+        names = set()
         if self.name:
-            names.append(self.name)
+            names.add(self.name)
+
+        names.update(self.options.get("names", []))
+        names.update(self.options.get("aliases", []))
+
+        if alias := self.options.get("alias", ""):
+            names.add(alias)
+
+        if self.is_pk():
+            names.add("pk")
+            if self.orm_class:
+                model_name = self.orm_class.model_name
+                names.add(f"{model_name}_id")
+                names.add(f"{model_name}_pk")
 
-        names.extend(self.options.get("names", []))
-        names.extend(self.options.get("aliases", []))
         return names
 
-    @classmethod
-    def get_instance(cls, **kwargs):
-        if not cls.instance:
-            cls.instance = cls(cls.type, cls.required, cls.options)
-        return cls.instance
+    @property
+    def interface_options(self):
+        if self.is_ref():
+            options = self.schema.pk.options
+            options.update(self.options)
+        else:
+            options = self.options
+        return options
 
     def __init__(self, field_type, field_required=False, field_options=None, **field_options_kwargs):
         """
@@ -645,18 +728,20 @@ class Field(object, metaclass=FieldMeta):
         # we aren't guaranteed to have this field's name when the descriptor is
         # created, so this will be the field name this descriptor will use to 
         # set the value onto the orm
-        self.orm_field_name = "_instance_{}_val".format(id(self))
+        #self.orm_field_name = "_instance_{}_val".format(id(self))
 
         # we keep a hash of the field's value when it was pulled from the
         # interface (see .iget) so we know if the field has been modified
-        self.orm_interface_hash = "_interface_{}_hash".format(id(self))
+        #self.orm_interface_hash = "_interface_{}_hash".format(id(self))
 
         field_options = utils.make_dict(field_options, field_options_kwargs)
 
         d = self.get_size(field_options)
         field_options.update(d)
 
-        self.orm_class = field_options.pop("orm_class", None)
+        #self.orm_class = field_options.pop("orm_class", None)
+        name = field_options.pop("name", "")
+        orm_class = field_options.pop("orm_class", None)
 
         choices = field_options.pop("choices", set())
         if choices or not self.choices:
@@ -671,29 +756,155 @@ class Field(object, metaclass=FieldMeta):
 
         self.set_type(field_type)
 
+        # this class is being created manually so mimic what the python parser
+        # would do
+        if name and orm_class:
+            self.__set_name__(orm_class, name)
+
+    def __set_name__(self, orm_class, name):
+        """This is called right after __init__
+
+        https://docs.python.org/3/howto/descriptor.html#customized-names
+
+        This is only called when an instance is created while a class is being
+        parsed/created, so if you just created an instance of Field you would
+        need to call this method manually
+
+        :param orm_class: type, the orm class this Field will belong to
+        :param name: str, the field's public name on the orm class
+        """
+        self.orm_class = orm_class
+        self.name = name
+
+        # the field name this descriptor will use to set the value onto the orm
+        # instance
+        self.orm_field_name = f"_{name}"
+
+        # we keep a hash of the field's value when it was pulled from the
+        # interface (see .iget) so we know if the field has been modified
+        self.orm_interface_hash = f"_interface_{name}_hash"
+
+        if orm_class:
+            logger.debug(
+                f"Field descriptor {orm_class.__name__}.{name} created with internal name {self.orm_field_name}"
+            )
+
     def get_size(self, field_options):
         """Figure out if this field has any size information"""
         d = {}
 
         min_size = field_options.pop("min_size", -1)
         max_size = field_options.pop("max_size", -1)
-        size = field_options.pop("size", -1)
+        size = field_options.pop("size", None)
 
-        if size > 0:
+        if size is not None:
             d['size'] = size
+
         else:
             if min_size > 0 and max_size < 0:
-                raise ValueError("min_size option was set with no corresponding max_size")
+                d['min_size'] = min_size
 
-            elif min_size < -1 and max_size > 0:
+            elif min_size < 0 and max_size > 0:
                 d['max_size'] = max_size
 
-            elif min_size >= 0 and max_size >= 0:
+            elif min_size > 0 and max_size > 0:
                 d['min_size'] = min_size
                 d['max_size'] = max_size
 
         return d
 
+    def size_info(self):
+        """Figure out actual sizing information with all the ways we can calculate
+        size now
+
+        :returns: dict, will always have "has_size" and "has_precision" keys
+            * if "has_size" key is True then "size" key will exist
+            * if "has_precision" is True then "precision" and "scale" keys will exist
+            * if "bounds" key exists it will be a tuple (min_size, max_size)
+        """
+        ret = {
+            "has_size": False,
+            "has_precision": False,
+            "original": {},
+        }
+        options = self.interface_options
+
+        if "size" in options:
+            ret["size"] = options["size"]
+            ret["original"]["size"] = options["size"]
+            ret["has_size"] = True
+            ret["bounds"] = (ret["size"], ret["size"])
+
+        if "max_size" in options:
+            ret["size"] = max(options["max_size"], ret.get("size", 0))
+            ret["original"]["max_size"] = options["max_size"]
+            ret["has_size"] = True
+            ret["bounds"] = (ret.get("size", 0), ret["size"])
+
+        if "min_size" in options:
+            ret["size"] = max(options["min_size"], ret.get("size", 0))
+            ret["original"]["min_size"] = options["min_size"]
+            ret["has_size"] = True
+            ret["bounds"] = (options["min_size"], ret["size"])
+
+        if ret["has_size"]:
+            # if size is like 15.6 then that would be considered 21
+            # precision with a scale of 6 (ie, you can have 15 digits before
+            # the decimal point and 6 after)
+            parts = str(ret["size"]).split(".")
+            if len(parts) > 1:
+                ret["scale"] = int(parts[1])
+                ret["precision"] = int(parts[0]) + ret["scale"]
+
+                # the set size was actually precision and scale so we don't have
+                # sizing information
+                ret["has_precision"] = True
+                ret.pop("size")
+
+        if "precision" in options:
+            # So the number 23.5141 has a precision of 6 and a scale of 4
+            precision = options["precision"]
+            ret["precision"] = int(precision)
+            ret["original"]["precision"] = precision
+            ret["has_precision"] = True
+            ret["scale"] = 0
+
+            if "scale" in options:
+                scale = options["scale"]
+                ret["scale"] = int(scale)
+                ret["original"]["scale"] = scale
+
+        interface_type = self.interface_type
+        if issubclass(interface_type, (int, float, decimal.Decimal)):
+            if "precision" in ret and "size" not in ret:
+                if ret["scale"]:
+                    ret["size"] = float("{}.{}".format(
+                        "9" * (ret["precision"] - ret["scale"]),
+                        "9" * ret["scale"],
+                    ))
+
+                else:
+                    ret["size"] = int("9" * ret["precision"])
+
+            elif "precision" not in ret and "size" not in ret:
+                # this is 32bit, it might be worth setting defined size to 64bit
+                ret["size"] = 2147483647
+                ret["precision"] = len(str(ret["size"]))
+
+            elif "precision" not in ret and "size" in ret:
+                ret["precision"] = len(str(ret["size"]))
+
+        elif issubclass(interface_type, (str, bytes, bytearray)):
+            if ret["has_size"]:
+                if not ret["has_precision"]:
+                    ret["precision"] = ret["size"]
+                    ret["scale"] = 0
+
+            elif ret["has_precision"]:
+                ret["size"] = ret["precision"]
+
+        return ret
+
     def set_type(self, field_type):
         """Try to infer as much about the type as can be inferred at this moment
 
@@ -706,27 +917,24 @@ class Field(object, metaclass=FieldMeta):
         :param field_type: mixed, the field type passed into __init__
         """
         std_types = (
+            str,
             bool,
-            long,
             int,
             float,
+            bytes,
             bytearray,
             decimal.Decimal,
             datetime.datetime,
             datetime.date,
-        )
-        if is_py2:
-            std_types = (basestring,) + std_types
-        else:
-            std_types = basestring + std_types
-
-        json_types = (
+            uuid.UUID,
             dict,
-            list,
         )
 
+        json_types = ()
+
         pickle_types = (
             set,
+            list,
         )
 
         self.original_type = field_type
@@ -774,6 +982,10 @@ class Field(object, metaclass=FieldMeta):
         """return True if this field is a primary key"""
         return self.options.get("pk", False)
 
+    def is_auto(self):
+        """Return True if this field auto-generates its value somehow"""
+        return self.options.get("auto", False)
+
     def is_ref(self):
         """return true if this field foreign key references the primary key of another orm"""
         return bool(self.schema)
@@ -794,6 +1006,9 @@ class Field(object, metaclass=FieldMeta):
         :returns: mixed
         """
         #pout.v("fget {}".format(self.name))
+        if self.is_ref():
+            # Foreign Keys get passed through their Field methods
+            val = self.schema.pk.fget(orm, val)
         return val
 
     def fgetter(self, v):
@@ -806,15 +1021,23 @@ class Field(object, metaclass=FieldMeta):
 
         think of this as when the orm receives the field value from the interface
 
-        :param orm: Orm
+        :param orm: Orm, the Orm instance the field is being set on. This can be
+            None if the select query had selected fields so a full orm instance isn't
+            being returned but rather just the selected values
         :param val: mixed, the current value of the field
         :returns: mixed
         """
         #pout.v("iget {}".format(self.name))
-        if self.is_serialized():
-            val = self.decode(val)
+        if self.is_ref():
+            # Foreign Keys get passed through their Field methods
+            val = self.schema.pk.iget(orm, val)
 
-        orm.__dict__[self.orm_interface_hash] = self.hash(orm, val)
+        else:
+            if self.is_serialized():
+                val = self.decode(val)
+
+            if orm:
+                orm.__dict__[self.orm_interface_hash] = self.hash(orm, val)
 
         return val
 
@@ -837,6 +1060,11 @@ class Field(object, metaclass=FieldMeta):
         if val is not None and self.choices:
             if val not in self.choices:
                 raise ValueError("Value {} not in {} value choices".format(val, self.name))
+
+        if self.is_ref():
+            # Foreign Keys get passed through their Field methods
+            val = self.schema.pk.fset(orm, val)
+
         return val
 
     def fsetter(self, v):
@@ -854,8 +1082,14 @@ class Field(object, metaclass=FieldMeta):
         :returns: mixed
         """
         #pout.v("iset {}".format(self.name))
-        if self.is_serialized():
-            val = self.encode(val)
+        if self.is_ref():
+            # Foreign Keys get passed through their Field methods
+            val = self.schema.pk.iset(orm, val)
+
+        else:
+            if self.is_serialized():
+                val = self.encode(val)
+
         return val
 
     def isetter(self, v):
@@ -935,6 +1169,9 @@ class Field(object, metaclass=FieldMeta):
         :param val: mixed, the fields value
         :returns: mixed
         """
+        if self.is_ref():
+            # Foreign Keys get passed through their Field methods
+            val = self.schema.pk.iquery(query, val)
         return val
 
     def iquerier(self, v):
@@ -943,13 +1180,19 @@ class Field(object, metaclass=FieldMeta):
         return self
 
     def jsonable(self, orm, val):
-        if val is None:
-            val = self.fdefault(orm, val)
+        if self.is_ref():
+            # Foreign Keys get passed through their Field methods
+            val = self.schema.pk.jsonable(orm, val)
+
+        else:
+            if val is None:
+                val = self.fdefault(orm, val)
+
+            if val is not None:
+                format_str = ""
+                if isinstance(val, (datetime.datetime, datetime.date)):
+                    val = Datetime(val).isoformat()
 
-        if val is not None:
-            format_str = ""
-            if isinstance(val, (datetime.datetime, datetime.date)):
-                val = Datetime(val).iso_8601()
         return val
 
     def jsonabler(self, v):
@@ -981,6 +1224,8 @@ class Field(object, metaclass=FieldMeta):
     def hash(self, orm, val):
         if self.is_serialized():
             return None
+        elif isinstance(val, dict):
+            return None
         return hash(val)
 
     def encode(self, val):
@@ -1017,7 +1262,7 @@ class Field(object, metaclass=FieldMeta):
 
         return val
 
-    def __get__(self, orm, classtype=None):
+    def __get__(self, orm, orm_class=None):
         """This is the wrapper that will actually be called when the field is
         fetched from the instance, this is a little different than Python's built-in
         @property fget method because it will pull the value from a shadow variable in
@@ -1055,3 +1300,88 @@ class Field(object, metaclass=FieldMeta):
         orm.__dict__[self.orm_field_name] = val
 
 
+class AutoUUID(Field):
+    """an auto-generating UUID field, by default this will be set as primary key"""
+    def __init__(self, **kwargs):
+        kwargs.setdefault("pk", True)
+        kwargs.setdefault("auto", True)
+        kwargs.setdefault("size", 36)
+        super().__init__(
+            field_type=uuid.UUID,
+            field_required=True,
+            **kwargs
+        )
+
+
+class AutoIncrement(Field):
+    """an auto-incrementing Serial field, by default this will be set as primary key"""
+    def __init__(self, **kwargs):
+        kwargs.setdefault("pk", True)
+        kwargs.setdefault("auto", True)
+        super().__init__(
+            field_type=int,
+            field_required=True,
+            **kwargs
+        )
+
+
+class AutoDatetime(Field):
+    """A special field that will create a datetime according to triggers
+
+    :Example:
+        # datetime will be auto-populated on creation
+        created = AutoDatetime(created=True)
+
+        # datetime will be auto-populated on creation and update
+        updated = AutoDatetime(updated=True)
+
+    The triggers are:
+        created: create a datetime in the field when the Orm is being created (added
+            to the db for the first time)
+        updated: update the datetime when the Orm is created and when it is updated
+            in the db
+    """
+    def __init__(self, **kwargs):
+        kwargs.setdefault("created", kwargs.get("create", True))
+        kwargs.setdefault("updated", kwargs.get("update", False))
+        kwargs.setdefault("auto", True)
+        if kwargs["updated"]:
+            kwargs["created"] = False
+
+        super().__init__(
+            field_type=datetime.datetime,
+            field_required=True,
+            **kwargs,
+        )
+
+    def iset(self, orm, val):
+        if self.options.get("updated", False):
+            return self.updated_iset(orm, val)
+
+        elif self.options.get("created", False):
+            return self.created_iset(orm, val)
+
+        else:
+            return super().iset(orm, val)
+
+    def created_iset(self, orm, val):
+        if not val and orm.is_insert():
+            #val = datetime.datetime.utcnow()
+            #val = datetime.datetime.now(datetime.timezone.utc)
+            val = Datetime()
+        return val
+
+    def updated_iset(self, orm, val):
+        if val:
+            if not self.modified(orm, val):
+                #val = datetime.datetime.utcnow()
+                #val = datetime.datetime.now(datetime.timezone.utc)
+                val = Datetime()
+
+        else:
+            #val = datetime.datetime.utcnow()
+            #val = datetime.datetime.now(datetime.timezone.utc)
+            val = Datetime()
+
+        return val
+
